/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "CLucene/StdHeader.h"
#include "KoreanFilter.h"
#include "KoreanTokenizer.h"
#include "morph/PatternConstants.h"

CL_NS_USE(analysis)
CL_NS_USE(util)
CL_NS_DEF2(analysis, kr)

const TCHAR* KoreanFilter::APOSTROPHE_TYPE = KoreanTokenizer::TOKEN_TYPES[KoreanTokenizer::APOSTROPHE];
const TCHAR* KoreanFilter::ACRONYM_TYPE    = KoreanTokenizer::TOKEN_TYPES[KoreanTokenizer::ACRONYM];

KoreanFilter::KoreanFilter(TokenStream *input, bool deleteTokenStream):
	TokenFilter(input, deleteTokenStream), bigrammable(true)
{
}

KoreanFilter::KoreanFilter(TokenStream *input, bool bigram, bool deleteTokenStream):
	TokenFilter(input, deleteTokenStream)
{
	bigrammable = bigram;
}

bool  KoreanFilter::next(Token *token) throw(MorphException) {
	if(koreanQueue.size()>0) {
		Token *t = koreanQueue.front();
		token->set(t->termText(), t->startOffset(), t->endOffset(), t->type());
		token->setPositionIncrement(t->getPositionIncrement());
		koreanQueue.delete_front();
		return true;
	}
			
	int skippedPositions = 0;
	bool hasToken;

	while (input->next(token)) {
		if (!_tcscmp(token->type(), KoreanTokenizer::TOKEN_TYPES[KoreanTokenizer::KOREAN])) {
			if (analysisKorean(token, skippedPositions)) {
				Token *t = koreanQueue.front();
				token->set(t->termText(), t->startOffset(), t->endOffset(), t->type());
				token->setPositionIncrement(t->getPositionIncrement());
				koreanQueue.delete_front();
			}
			else {
				skippedPositions++;			    		
				continue;
			}
		}
		else {
			if (!analysisETC(token)) {
				skippedPositions++;			    		
				continue;
			}
		}
		return true;
	}

	return false;
}
	
bool KoreanFilter::analysisKorean(Token *token, int skipinc) throw(MorphException) {
	const TCHAR* input = token->termText();
	AnalysisOutputList outputs;
	morph.analyze(input, outputs);
		
	HashMap map;
	for (AnalysisOutputList::iterator output = outputs.begin();
		 output != outputs.end(); output++)
	{
		if (output->getPos()==PatternConstants::POS_NOUN) {
			map[output->getStem()] = 1;				
			if (output->getScore() == AnalysisOutput::SCORE_CORRECT) {
				CompoundEntryList cnouns = output->getCNounList();				
				for(CompoundEntryList::iterator cnoun = cnouns.begin();
					cnoun != cnouns.end(); cnoun++) {
					if(cnoun->getWord().length()>1) map[cnoun->getWord()] = 0;
				}
			} else if(bigrammable) {
				addBiagramToMap(output->getStem(),map);
			}
		}
	}
		
	HashMap::iterator iter;
	int i = 0;
	for (iter = map.begin(); iter != map.end(); iter++) {
		const String &text = iter->first;
		int index = _tcsstr(input, text.c_str()) - input;
		if (index == (int)input) index = -1;
		Token *t = new Token(text.c_str(),
							 token->startOffset()+(index!=-1 ? index : 0),
							 index!=-1 ? token->startOffset()+index+text.length() : token->endOffset(),
							 KoreanTokenizer::TOKEN_TYPES[KoreanTokenizer::KOREAN]);
		koreanQueue.push_back(t);
		if (i == 0) t->setPositionIncrement(token->getPositionIncrement()+skipinc);
		else t->setPositionIncrement(0);
		i++;
	}
		
	if(koreanQueue.size()==0) return false;

	return true;
}
	
void KoreanFilter::addBiagramToMap(const String& input, HashMap& map) {
	int offset = 0;
	int strlen = input.length();
	while(offset<strlen-1) {
		if(isAlphaNumChar(input.charAt(offset))) {
			String text = findAlphaNumeric(input.substring(offset));
			map[text] = 0;
			offset += text.length();
		} else {
			String text = input.substring(offset,
										  offset+2>strlen?strlen:offset+2);
			map[text] = 0;
			offset++;
		}				
	}
}
	
String KoreanFilter::findAlphaNumeric(const String& text) {
	int pos = 0;
	for(int i=0;i<text.length();i++) {
		if(!isAlphaNumChar(text.charAt(i))) break;
		pos++;
	}				
	return text.substring(0,pos);
}
	
bool KoreanFilter::analysisETC(Token *t) throw(MorphException) {
	TCHAR *buffer = t->_termText;
	const int bufferLength = t->termTextLength();
	const TCHAR *type = t->type();

	if (type == APOSTROPHE_TYPE &&		  // remove 's
		bufferLength >= 2 &&
		buffer[bufferLength-2] == _T('\'') &&
		(buffer[bufferLength-1] == _T('s') || buffer[bufferLength-1] == _T('S')) ) {
		// Strip last 2 characters off
		buffer[bufferLength-2]=0; 
		t->resetTermTextLen();
	}
	else if (type == ACRONYM_TYPE) {		  // remove dots
		int upto = 0;
		for(int i=0;i<bufferLength;i++) {
	        TCHAR c = buffer[i];
	        if (c != _T('.'))
				buffer[upto++] = c;
		}
		buffer[upto] = 0;
		t->resetTermTextLen();
	}

	return true;
}
	
bool KoreanFilter::isAlphaNumChar(TCHAR c) {
	if((c>=48&&c<=57)||(c>=65&&c<=122)) return true;		
	return false;
}


CL_NS_END2
