/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "CLucene/StdHeader.h"
#include "KoreanTokenizer.h"

CL_NS_USE(analysis)
CL_NS_USE(util)
CL_NS_DEF2(analysis, kr)
	
const TCHAR* KoreanTokenizer::TOKEN_TYPES[] = {
		_T("<ALPHANUM>"),
		_T( "<APOSTROPHE>"),
	    _T("<ACRONYM>"),
	    _T("<COMPANY>"),
	    _T("<EMAIL>"),
	    _T("<HOST>"),
	    _T("<NUM>"),
		_T("<CJ>"),
		_T("<ACRONYM_DEP>"),
		_T("<KOREAN>")
};

KoreanTokenizer::KoreanTokenizer(CL_NS(util)::Reader* input):
	Tokenizer(input)
{
	bufferLength = DEFAULT_BUFLEN;
	buffer = new TCHAR[bufferLength];
	bufferEnd = buffer+bufferLength;
	bufferOffset = 0;
	
	limit  = buffer;
	cursor = buffer;
	start  = 0;
	marker = 0;
}

KoreanTokenizer::~KoreanTokenizer()
{
	delete[] buffer;
}

bool KoreanTokenizer::next(Token* token)
{
	// null check of token parameter is needed?
	
	int tokenType;

	#define YYCTYPE         TCHAR
	#define YYCURSOR        cursor
	#define YYLIMIT         limit
	#define YYMARKER        marker
	#define YYFILL(n)		yyfill()

	for (;;)
	{
	start = cursor;		// save token start position
	
/*!re2c
  /**/
	LETTER = [\u0041-\u005a\u0061-\u007a\u00c0-\u00d6\u00d8-\u00f6\u00f8-\u00ff\u0100-\u1fff\uffa0-\uffdc];

	DIGIT = [\u0030-\u0039\u0660-\u0669\u06f0-\u06f9\u0966-\u096f\u09e6-\u09ef\u0a66-\u0a6f\u0ae6-\u0aef\u0b66-\u0b6f\u0be7-\u0bef\u0c66-\u0c6f\u0ce6-\u0cef\u0d66-\u0d6f\u0e50-\u0e59\u0ed0-\u0ed9\u1040-\u1049];

	HANLETTER = [\uac00-\ud7af\u1100-\u11ff];

	/* Chinese, Japanese */
	CJ = [\u3040-\u318f\u3100-\u312f\u3040-\u309F\u30A0-\u30FF\u31F0-\u31FF\u3300-\u337f\u3400-\u4dbf\u4e00-\u9fff\uf900-\ufaff\uff65-\uff9f];

	WHITESPACE = "\r\n" | [ \r\n\t\f];

	/* punctuation */
	P = ("_"|"-"|"/"|"."|",");

	ALPHA = (LETTER)+;

	/* basic word: a sequence of digits & letters */
	ALPHANUM = (LETTER | DIGIT)+;

	HAS_DIGIT = (LETTER | DIGIT)* DIGIT (LETTER | DIGIT)*;
	
	/* floating point, serial, model numbers, ip addresses, etc. */
	/* every other segment must have at least one digit			 */
	NUM = "." DIGIT	| DIGIT "."
		| (ALPHA P)* HAS_DIGIT (P ALPHANUM)+;
	
	 /* korean word: a sequence of digits & letters & */
	KOREAN = (HANLETTER | LETTER+ HANLETTER+ | NUM+ HANLETTER+ | DIGIT+ HANLETTER+ )+;

	/* internal apostrophes: O'Reilly, you're, O'Reilly's */
	/* use a post-filter to remove possesives			  */
	APOSTROPHE = ALPHA ("'" ALPHA)+;

	/* acronyms: U.S.A., I.B.M., etc.	*/
	/* use a post-filter to remove dots */
	ACRONYM = LETTER "." (LETTER ".")+;

	ACRONYM_DEP	= ALPHANUM "." (ALPHANUM ".")+;

	/* company names like AT&T and Excite@Home. */
	COMPANY = ALPHA ("&"|"@") ALPHA;

	/* email addresses */
	EMAIL = ALPHANUM (("."|"-"|"_") ALPHANUM)* "@" ALPHANUM (("."|"-") ALPHANUM)+;

	/* hostname */
	HOST = ALPHANUM ("." ALPHANUM)+;

	KOREAN				{ tokenType = KOREAN;		goto SET_TOKEN; }
	ALPHANUM			{ tokenType = ALPHANUM;		goto SET_TOKEN; }
	APOSTROPHE			{ tokenType = APOSTROPHE;	goto SET_TOKEN; }
	ACRONYM				{ tokenType = ACRONYM;		goto SET_TOKEN; }
	COMPANY				{ tokenType = COMPANY;		goto SET_TOKEN; }
	EMAIL				{ tokenType = EMAIL;		goto SET_TOKEN; }
	HOST				{ tokenType = HOST;			goto SET_TOKEN; }
	NUM					{ tokenType = NUM;			goto SET_TOKEN; }
	CJ					{ tokenType = CJ;			goto SET_TOKEN; }
	ACRONYM_DEP			{ tokenType = ACRONYM_DEP;	goto SET_TOKEN; }
	'\u0000'			{ return false; }
	[^]					{ continue; }
*/

SET_TOKEN:
	int posIncr = 1;
	
	if (cursor-start <= maxTokenLength) {
		int32_t tokenLength = cursor-start;
		int32_t startOffset = bufferOffset+start-buffer;

		_tcsncpy(tokenBuffer, start, tokenLength);
		tokenBuffer[tokenLength] = '\0';		
		token->setText(tokenBuffer);
		token->setPositionIncrement(posIncr);
		token->setStartOffset(startOffset);
		token->setEndOffset(startOffset+tokenLength);
		
		// This 'if' should be removed in the next release. For now, it converts
		// invalid acronyms to HOST. When removed, only the 'else' part should
		// remain.
		if (tokenType == ACRONYM_DEP) {
			token->setType(TOKEN_TYPES[ACRONYM]);
		} else {
			token->setType(TOKEN_TYPES[tokenType]);
		}
		return true;
	} else
		// When we skip a too-long term, we still increment the
		// position increment
		posIncr++;
	
	} /* for(;;) */
}

void KoreanTokenizer::yyfill()
{
    /* first: make room (if you can) */
    if (start > buffer) {
		int32_t startOffset = start - buffer;
		memmove(buffer, start, (limit-start)*sizeof(TCHAR));
		marker -= startOffset;
		cursor -= startOffset;
		limit -= startOffset;
		start -= startOffset;
		bufferOffset += startOffset;
    }

	/* is the buffer big enough? */
    if (cursor >= bufferEnd) {
      /* if not: blow it up */
		int32_t limitOffset = limit-buffer;
		int32_t startOffset = start-buffer;
		int32_t markerOffset = marker-buffer;
		int32_t cursorOffset = cursor-buffer;
			
		bufferLength = bufferLength*2;
		TCHAR *newBuffer = new TCHAR[bufferLength];
		_tcsncpy(newBuffer, buffer, limit-start);
		delete buffer;
		buffer = newBuffer;
		bufferEnd = buffer+bufferLength;

		marker = markerOffset + buffer;
		cursor = cursorOffset + buffer;
		start = buffer + startOffset;
		limit = limitOffset + buffer;
    }    

    /* finally: fill the buffer with new input */
	const TCHAR *ioBuffer = NULL;
    int numRead = input->read(ioBuffer, bufferEnd-limit);
	
    if (numRead < 0) {
		*(limit++) = '\0';	/* end of stream */
		return;
	}
	_tcsncpy(limit, ioBuffer, numRead);
	limit += numRead;
}

CL_NS_END2
