#ifndef _lucene_analysis_kr_Trie
#define _lucene_analysis_kr_Trie

#include "CLucene/StdHeader.h"
#include "utils/String.h"
#include "UnmodifiableIterator.h"

CL_NS_DEF2(analysis, kr)

class NoSuchElementException {};

template <typename V> class TrieNode;

/**
 * A labelled edge, i.e., a String label and a TrieNode endpoint.
 */
template<typename V>
class TrieEdge {
private:
    String label;
    TrieNode<V> *child;

public:
    /**
     * @requires label.size() > 0
     * @requires child != null
     */
    TrieEdge(const String &label, TrieNode<V> *_child):
		label(label), child(_child) { }	

	void clearAll() { child->clearAll(); delete child; }	
	
    const String& getLabel() const {
        return label;
    }

    /**
     * Returns the first character of the label, i.e., getLabel().charAt(0).
     */
    TCHAR getLabelStart() const {
        // You could store this char as an optimization if needed.
        return label.charAt(0);
    }

    TrieNode<V>* getChild() {
        return child;
    }

};

/**
 * A node of the Trie. Each Trie has a list of children, labelled by strings.
 * Each of these [String label, TrieNode child] pairs is considered an "edge".
 * The first character of each label must be distinct. When managing children,
 * different implementations may trade space for time. Each node also stores an
 * arbitrary Object value.
 * <p>
 * 
 * Design note: this is a "dumb" class. It is <i>only</i> responsible for
 * managing its value and its children. None of its operations are recursive;
 * that is Trie's job. Nor does it deal with case.
 */
template <typename V>
class TrieNode {
public:
	typedef CL_NS(util)::CLVector<TrieEdge<V>*, CL_NS(util)::Deletor::Void<TrieEdge<V> > > EdgeList;
	
private:
    /**
     * The value of this node.
     */
    V *value;

    /**
     * The list of children. Children are stored as a sorted Vector because it
     * is a more compact than a tree or linked lists. Insertions and deletions
     * are more expensive, but they are rare compared to searching.
     * <p>
     * 
     * INVARIANT: children are sorted by distinct first characters of edges,
     * i.e., for all i &lt; j,<br>
     * children[i].edge->charAt(0) &lt; children[j].edge->charAt(0)
     */
    EdgeList children;

public:
    TrieNode() : value(NULL) {}
    TrieNode(V *v) : value(v) {}	

	~TrieNode() { if (value != NULL) delete value; }

	void clearAll() {
		if (value != NULL) { delete value; value = NULL; }
		for (int i = 0; i < children.size(); i++)
			children[i]->clearAll();
		children.clear();
	}
	
    V* getValue() {	return value; }
	
    void setValue(V *v) {
		if (value != NULL) delete value;
		value = v;
	}

private:	
    /**
     * Get the nth child edge of this node.
     * 
     * @requires 0 &lt;= i &lt; children.size()
     */
    TrieEdge<V>* get(int i) {
        return children.at(i);
    }

    /**
     * (internal) If exact, returns the unique i so that:
     * children[i].getLabelStart() == c<br>
     * If !exact, returns the largest i so that: children[i].getLabelStart()
     * &lt;= c<br>
     * In either case, returns -1 if no such i exists.
     * <p>
     * 
     * This method uses binary search and runs in O(log N) time, where N =
     * children.size().<br>
     * The standard Java binary search methods could not be used because they
     * only return exact matches. Also, they require allocating a dummy Trie.
     * 
     * Example1: Search non exact c == '_' in {[0] => 'a...', [1] => 'c...'};
     * start loop with low = 0, high = 1; middle = 0, cmiddle == 'a', c <
     * cmiddle, high = 0 (low == 0); middle = 0, cmiddle == 'a', c < cmiddle,
     * high = -1 (low == 0); end loop; return high == -1 (no match, insert at
     * 0). Example2: Search non exact c == 'a' in {[0] => 'a', [1] => 'c'} start
     * loop with low = 0, high = 1; middle = 0, cmiddle == 'a', c == cmiddle,
     * abort loop by returning middle == 0 (exact match). Example3: Search non
     * exact c == 'b' in {[0] => 'a...', [1] => 'c...'}; start loop with low =
     * 0, high = 1; middle = 0, cmiddle == 'a', cmiddle < c, low = 1 (high ==
     * 1); middle = 1, cmiddle == 'c', c < cmiddle, high = 0 (low == 1); end
     * loop; return high == 0 (no match, insert at 1). Example4: Search non
     * exact c == 'c' in {[0] => 'a...', [1] => 'c...'}; start loop with low =
     * 0, high = 1; middle = 0, cmiddle == 'a', cmiddle < c, low = 1 (high ==
     * 1); middle = 1, cmiddle == 'c', c == cmiddle, abort loop by returning
     * middle == 1 (exact match). Example5: Search non exact c == 'd' in {[0] =>
     * 'a...', [1] => 'c...'}; start loop with low = 0, high = 1; middle = 0,
     * cmiddle == 'a', cmiddle < c, low = 1 (high == 1); middle = 1, cmiddle ==
     * 'c', cmiddle < c, low = 2 (high == 1); end loop; return high == 1 (no
     * match, insert at 2).
     */
    int search(TCHAR c, bool exact) {
        // This code is stolen from IntSet.search.
        int low = 0;
        int high = children.size() - 1;
        while (low <= high) {
            int middle = (low + high) / 2;
            TCHAR cmiddle = get(middle)->getLabelStart();
            if (cmiddle < c)
                low = middle + 1;
            else if (c < cmiddle)
                high = middle - 1;
            else
                // c == cmiddle
                return middle; // Return exact match.
        }
        if (exact)
            return -1; // Return no match.
        return high; // Return closest *lower or equal* match. (This works!)
    }

    /**
     * Returns the edge (at most one) whose label starts with the given
     * character, or null if no such edge.
     */
public:
	TrieEdge<V>* get(TCHAR labelStart) {
        int i = search(labelStart, true);
        if (i < 0)
            return NULL;
        return get(i);
    }

    /**
     * Inserts an edge with the given label to the given child to this. Keeps
     * all edges binary sorted by their label start.
     * 
     * @requires label not empty.
     * @requires for all edges E in this, label.getLabel[0] != E not already
     *           mapped to a node.
     * @modifies this
     */
    void put(const String &label, TrieNode<V> *child) {
        int i;
        // If there's a match it is the closest lower or equal one, and
        // precondition requires it to be lower, so we add the edge *after*
        // it. If there's no match, there are two cases: the Trie is empty,
        // or the closest match returned is the last edge in the list.
        if ((i = search(label.charAt(0), false)) >= 0) // find closest match
		{
        }
        children.insert(children.begin()+(i+1), new TrieEdge<V>(label, child));
    }

    /**
     * Removes the edge (at most one) whose label starts with the given
     * character. Returns true if any edges where actually removed.
     */
    bool remove(TCHAR labelStart) {
        int i;
        if ((i = search(labelStart, true)) < 0)
            return false;
        children.remove(i);
        return true;
    }

    /**
     * Ensures that this's children take a minimal amount of storage. This
     * should be called after numerous calls to add().
     * 
     * @modifies this
     */
    void trim() {
		// FIXME: not verified code
		// EdgeList(children, false).swap(children);
    }

    /**
     * Maps (lambda(edge) edge->getChild) on children.iterator().
     */
	class ChildrenForwardIterator : public UnmodifiableIterator<V> {
		TrieNode<V> *node;
        int i;

	public:
		ChildrenForwardIterator(TrieNode<V> *_node): node(_node), i(0) {}
        bool hasNext() {
            return i < node->children.size();
        }

        TrieNode<V>* next() {
            if (i < node->children.size())
                return node->get(i++)->getChild();
			throw NoSuchElementException();
        }
    };

	/**
     * Returns the children of this in forward order, as an iterator of
     * TrieNode.
     */
	ChildrenForwardIterator childrenForward() {
        return ChildrenForwardIterator(this);
    }
	
    /**
     * Returns the children of this in forward order, as an iterator of
     * TrieNode.
     */
    /*
     * public: Iterator childrenBackward() { return new
     * ChildrenBackwardIterator(); }
     */

    /**
     * Maps (lambda(edge) edge->getChild) on children.iteratorBackward().
     */
    /*
     * private: class ChildrenBackwardIterator extends UnmodifiableIterator { int
     * i = children.size() - 1;
     * 
     * public: bool hasNext() { return i >= 0; }
     * 
     * public: Object next() { if (i >= 0) return get(i--).getChild(); throw new
     * NoSuchElementException(); } }
     */

    /**
     * Returns the labels of the children of this in forward order, as an
     * iterator of Strings.
     */

    /**
     * Maps (lambda(edge) edge->getLabel) on children.iterator()
     */
	class LabelForwardIterator : public UnmodifiableIterator<V> {
		TrieNode<V> *node;
        int i;

	public:
		LabelForwardIterator(TrieNode<V> *_node): node(_node), i(0) {}
		
        bool hasNext() {
            return i < node->children.size();
        }

        const V& next() {
            if (i < node->children.size())
                return node->get(i++).getLabel();
            throw NoSuchElementException();
        }
    };

	LabelForwardIterator labelsForward() {
		return LabelForwardIterator(this);
    }
	
    /**
     * Returns the labels of the children of this in backward order, as an
     * iterator of Strings.
     */
    /*
     * public: Iterator labelsBackward() { return new LabelBackwardIterator(); }
     */

    /**
     * Maps (lambda(edge) edge->getLabel) on children.iteratorBackward()
     */
    /*
     * private: class LabelBackwardIterator extends UnmodifiableIterator { int i =
     * children.size() - 1;
     * 
     * public: bool hasNext() { return i >= 0; }
     * 
     * public: Object next() { if (i >= 0) return get(i--).getLabel(); throw new
     * NoSuchElementException(); } }
     */

//     // inherits javadoc comment.
//     public: String toString() {
//         Object val = getValue();
//         if (val != NULL)
//             return val.toString();
//         return "NULL"; //$NON-NLS-1$
//     }

    /**
     * Unit test.
     * 
     * @see TrieNodeTest
     */
};


/*
 * An information reTRIEval tree, a.k.a., a prefix tree. A Trie is similar to a
 * dictionary, except that keys must be strings. Furthermore, Trie provides an
 * efficient means (getPrefixedBy()) to find all values given just a PREFIX of a
 * key.
 * <p>
 * 
 * All retrieval operations run in O(nm) time, where n is the size of the
 * key/prefix and m is the size of the alphabet. Some implementations may reduce
 * this to O(n log m) or even O(n) time. Insertion operations are assumed to be
 * infrequent and may be slower. The space required is roughly linear with
 * respect to the sum of the sizes of all keys in the tree, though this may be
 * reduced if many keys have common prefixes.
 * <p>
 * 
 * The Trie can be set to ignore case. Doing so is the same as making all keys
 * and prefixes lower case. That means the original keys cannot be extracted
 * from the Trie.
 * <p>
 * 
 * Restrictions (not necessarily limitations!)
 * <ul>
 * <li><b>This class is not synchronized.</b> Do that externally if you
 * desire.
 * <li>Keys and values may not be null.
 * <li>The interface to this is not complete.
 * </ul>
 * 
 * See http://www.csse.monash.edu.au/~lloyd/tildeAlgDS/Tree/Trie.html for a
 * discussion of Tries.
 */

template<typename S, typename V>
class Trie {
public:
	// FIXME: Now useless, just returning pointer is better.
	class Position {
		V *value;
	public:
		Position(V *v = NULL):value(v) {}
		V& operator*() const { return *value; }
		V* operator->() const { return value; }
		bool isNull() const { return (value == NULL) ? true : false; }
	};
	
    /**
     * Our representation consists of a tree of nodes whose edges are labelled
     * by strings. The first characters of all labels of all edges of a node
     * must be distinct. Typically the edges are sorted, but this is determined
     * by TrieNode.
     * <p>
     * 
     * An abstract TrieNode is a mapping from String keys to values, { <K1, V1>,
     * ..., <KN, VN> }, where all Ki and Kj are distinct for all i != j. For any
     * node N, define KEY(N) to be the concatenation of all labels on the edges
     * from the root to that node. Then the abstraction function is:
     * <p>
     * 
     * <blockquote> { <KEY(N), N.getValue() | N is a child of root and
     * N.getValue() != null} </blockquote>
     * 
     * An earlier version used character labels on edges. This made
     * implementation simpler but used more memory because one node would be
     * allocated to each character in long strings if that string had no common
     * prefixes with other elements of the Trie.
     * <p>
     * 
     * <dl>
     * <dt>INVARIANT:</td>
     * <dd>For any node N, for any edges Ei and Ej from N,<br>
     * i != j &lt;==&gt; Ei.getLabel().getCharAt(0) !=
     * Ej.getLabel().getCharAt(0)</dd>
     * <dd>Also, all invariants for TrieNode and TrieEdge must hold.</dd>
     * </dl>
     */
private:
    TrieNode<V> *root;

    /**
     * Indicates whever search keys are case-sensitive or not. If true, keys
     * will be canonicalized to lowercase.
     */
    bool ignoreCase;

public:
    /**
     * Constructs a new, empty tree.
     */
	Trie(bool ignoreCase) {
        this->ignoreCase = ignoreCase;
		root = new TrieNode<V>();
    }

	~Trie() { clear(); }
	
    /**
     * Makes this empty.
     * 
     * @modifies this
     */
    void clear() {
		if (root != NULL) {
			root->clearAll();
			delete root;
		}
    }
	void reset() { clear(); root = new TrieNode<V>(); };

public:
    /**
     * Returns the canonical version of the given string.
     * <p>
     * 
     * In the basic version, strings are added and searched without
     * modification. So this simply returns its parameter s.
     * <p>
     * 
     * Other overrides may also perform a conversion to the NFC form
     * (interoperable across platforms) or to the NFKC form after removal of
     * accents and diacritics from the NFKD form (ideal for searches using
     * strings in natural language).
     * <p>
     * 
     * Made public: instead of protected, because the public: Prefix operations
     * below may need to use a coherent conversion of search prefixes.
     */
	// FIXME: \u00df -> \u0053 \u0053
	void canonicalCase(String &s) {
        if (ignoreCase)			
			s.casefold();
    }
	
private:
    /**
     * Matches the pattern <tt>b</tt> against the text
     * <tt>a[startOffset...stopOffset - 1]</tt>.
     * 
     * @return the first <tt>j</tt> so that:<br>
     *         <tt>0 &lt;= i &lt; b.length()</tt> AND<br>
     *         <tt>a[startOffset + j] != b[j]</tt> [a and b differ]<br>
     *         OR <tt>stopOffset == startOffset + j</tt> [a is undefined];<br>
     *         Returns -1 if no such <tt>j</tt> exists, i.e., there is a
     *         match.<br>
     *         Examples:
     *         <ol>
     *         <li>a = "abcde", startOffset = 0, stopOffset = 5, b = "abc"<br>
     *         abcde ==&gt; returns -1<br>
     *         abc
     *         <li>a = "abcde", startOffset = 1, stopOffset = 5, b = "bXd"<br>
     *         abcde ==&gt; returns 1 bXd
     *         <li>a = "abcde", startOffset = 1, stopOffset = 3, b = "bcd"<br>
     *         abc ==&gt; returns 2<br>
     *         bcd
     *         </ol>
     * 
     * @requires 0 &lt;= startOffset &lt;= stopOffset &lt;= a.length()
     */
	const int match(const String& a, int startOffset, int stopOffset, const String& b) {
        // j is an index into b
        // i is a parallel index into a
        int i = startOffset;
        for (int j = 0; j < b.length(); j++) {
            if (i >= stopOffset)
                return j;
            if (a.charAt(i) != b.charAt(j))
                return j;
            i++;
        }
        return -1;
    }

public:
    /**
     * Maps the given key (which may be empty) to the given value.
     * 
     * @return the old value associated with key, or <tt>null</tt> if none
     * @requires value != null
     * @modifies this
     */
	bool add(const String& key, V* value) {
		return add(key, value, NULL);
	}

	bool add(const String& _key, V* value, V** /*out*/ past) {
        // early conversion of key, for best performance
		String key(_key);
        canonicalCase(key);
        // Find the largest prefix of key, key[0..i - 1], already in this->
        TrieNode<V> *node = root;
        int i = 0;
        while (i < key.length()) {
            // Find the edge whose label starts with key[i].
            TrieEdge<V> *edge = node->get(key.charAt(i));
            if (edge == NULL) {
                // 1) Additive insert.
				TrieNode<V> *newNode = new TrieNode<V>(value);
                node->put(key.substring(i), newNode);
                return false;
            }
            // Now check that rest of label matches
            const String &label = edge->getLabel();
            int j = match(key, i, key.length(), label);
            if (j >= 0) {
                TrieNode<V> *child = edge->getChild();
                TrieNode<V> *intermediate = new TrieNode<V>();
                String a = label.substring(0, j);
                // Assert.that(canonicalCase(a).equals(a), "Bad edge a");
                String b = label.substring(j);
                // Assert.that(canonicalCase(b).equals(b), "Bad edge a");
                String c = key.substring(i + j);
                if (c.length() > 0) { // Split.
                    TrieNode<V> *newNode = new TrieNode<V>(value);
                    node->remove(label.charAt(0));
					intermediate->put(b, child);
                    intermediate->put(c, newNode);
                    node->put(a, intermediate);                    
                } else { // Splice.
                    node->remove(label.charAt(0));
					intermediate->put(b, child);
                    intermediate->setValue(value);
                    node->put(a, intermediate);                    
                }
                return false;
            }
            // Prefix overlaps perfectly with all of edge label.
            // Keep searching.
            node = edge->getChild();
            i += label.length();
        }
        // 3) Relabel insert. Prefix already in this, though not necessarily
        // associated with a value.
		if (past != NULL) *past = new V(*node->getValue());
        node->setValue(value);
        return true;
    }

    /**
     * Returns the node associated with prefix, or null if none. (internal)
     */
private:
	TrieNode<V>* fetch(const String &prefix) {
        // This private: method uses prefixes already in canonical form.
        TrieNode<V> *node = root;
        for (int i = 0; i < prefix.length();) {
            // Find the edge whose label starts with prefix[i].
            TrieEdge<V> *edge = node->get(prefix.charAt(i));
            if (edge == NULL)
                return NULL;
            // Now check that rest of label matches.
            const String &label = edge->getLabel();
            int j = match(prefix, i, prefix.length(), label);
            if (j != -1)
                return NULL;
            i += label.length();
            node = edge->getChild();
        }
        return node;
    }

    /**
     * Returns the value associated with the given key, or null if none.
     * 
     * @return the <tt>Object</tt> value or <tt>null</tt>
     */
public:
	Position find(const String &_key) {
		String key(_key);
        // early conversion of search key
        canonicalCase(key);
        // search the node associated with key, if it exists
        TrieNode<V> *node = fetch(key);
        if (node == NULL)
            return Position(NULL);
        // key exists, return the value
		return Position(node->getValue());
    }

    /**
     * Ensures no values are associated with the given key.
     * 
     * @return <tt>true</tt> if any values were actually removed
     * @modifies this
     */
    bool remove(const String& _key) {
		String key(_key);
        // early conversion of search key		
        canonicalCase(key);
        // search the node associated with key, if it exists
        TrieNode<V> *node = fetch(key);
        if (node == NULL)
            return false;
        // key exists and can be removed.
        // TODO: prune unneeded nodes to save space
        bool ret = node->getValue() != NULL;
        node.setValue(NULL);
        return ret;
    }

	class NodeIterator : public UnmodifiableIterator<TrieNode<V> *> {
        /**
         * Stack for DFS. Push and pop from back. The last element of stack is
         * the next node who's value will be returned.
         * <p>
         * 
         * INVARIANT: Top of stack contains the next node with not null value to
         * pop. All other elements in stack are iterators.
         */
	private:
		typedef typename TrieNode<V>::ChildrenForwardIterator ChildrenForwardIterator;
		typedef CL_NS(util)::CLVector<ChildrenForwardIterator,
						  CL_NS(util)::Deletor::NullVal<ChildrenForwardIterator> > IteratorList;
		IteratorList stack;
        bool withNulls;
		TrieNode<V> *nextNode;

	public:
        /**
         * Creates a new iterator that yields all the nodes of start and its
         * children that have values (ignoring internal nodes).
         */
        NodeIterator(TrieNode<V> *start, bool _withNulls):
			nextNode(NULL), withNulls(_withNulls) {
			if (start != NULL) {
				if (withNulls || start->getValue() != NULL)
					// node has a value, push it for next
					nextNode = start;
				else
					// scan node children to find the next node
					advance(start);
			}
        }

        // inherits javadoc comment
        bool hasNext() {
            return (nextNode != NULL);
        }

        // inherits javadoc comment
        TrieNode<V>* next() {
            int size;
			TrieNode<V> *node = nextNode;			
            if (nextNode == NULL)
                throw NoSuchElementException();
            advance(node);
            return node;
        }

	private:
        /**
         * Scan the tree (top-down) starting at the already visited node until
         * finding an appropriate node with not null value for next(). Keep
         * unvisited nodes in a stack of siblings iterators. Return either an
         * empty stack, or a stack whose top will be the next node returned by
         * next().
         */
		void advance(TrieNode<V> *node) {
			ChildrenForwardIterator children = node->childrenForward();
            while (true) { // scan siblings and their children
                int size;
                if (children.hasNext()) {
                    node = children.next();
                    if (children.hasNext()) // save siblings
                        stack.push_back(children);
                    // check current node and scan its sibling if necessary
                    if (withNulls || node->getValue() == NULL)
                        children = node->childrenForward(); // loop from there
                    else { // node qualifies for next()
                        nextNode = node;
                        return; // next node exists
                    }
                } else if ((size = stack.size()) == 0) {
					this->nextNode= NULL;
                    return; // no next node
				}
                else {
                    // no more siblings, return to parent
                    children = stack.back();
					stack.pop_back();
				}
            }
        }
    };
		

    /**
     * Returns all the (non-null) values associated with a given node and its
     * children. (internal)
     */
	class ValueIterator : public NodeIterator {
	public:
        ValueIterator(TrieNode<V> *start) :
			NodeIterator(start, false) { }
	   
		// inherits javadoc comment
		V* next() {
            return NodeIterator::next()->getValue();
        }
    };

	
    /**
     * Returns an iterator (of Object) of the values mapped by keys in this that
     * start with the given prefix, in any order. That is, the returned iterator
     * contains exactly the values v for which there exists a key k so that
     * k.startsWith(prefix) and get(k) == v. The remove() operation on the
     * iterator is unimplemented.
     * 
     * @requires this not modified while iterator in use
     */
    ValueIterator getPrefixedBy(const String& _prefix) {
		String prefix(_prefix);
        // Early conversion of search key
        canonicalCase(prefix);
        // Note that canonicalization MAY have changed the prefix length!
        return getPrefixedBy(prefix, 0, prefix.length());
    }

    /**
     * Same as getPrefixedBy(prefix.substring(startOffset, stopOffset). This is
     * useful as an optimization in certain applications to avoid allocations.
     * <p>
     * 
     * Important: canonicalization of prefix substring is NOT performed here!
     * But it can be performed early on the whole buffer using the public: method
     * <tt>canonicalCase(String)</tt> of this->
     * 
     * @requires 0 &lt;= startOffset &lt;= stopOffset &lt;= prefix.length
     * @see #canonicalCase(String)
     */
    ValueIterator getPrefixedBy(const String& prefix, int startOffset, int stopOffset) {
        // Find the first node for which "prefix" prefixes KEY(node). (See the
        // implementation overview for a definition of KEY(node).) This code is
        // similar to fetch(prefix), except that if prefix extends into the
        // middle of an edge label, that edge's child is considered a match.
        TrieNode<V> *node = root;
        for (int i = startOffset; i < stopOffset;) {
            // Find the edge whose label starts with prefix[i].
            TrieEdge<V> *edge = node->get(prefix.charAt(i));
            if (edge == NULL) {
                return ValueIterator(NULL);
            }
            // Now check that rest of label matches
            node = edge->getChild();
            String label = edge->getLabel();
            int j = match(prefix, i, stopOffset, label);
            if (i + j == stopOffset) {
                // a) prefix overlaps perfectly with just part of edge label
                break;
            } else if (j >= 0) {
                // b) prefix and label differ at some point
                node = NULL;
                break;
            } else {
                // c) prefix overlaps perfectly with all of edge label.
            }
            i += label.length();
        }
        // Yield all children of node, including node itself.
        if (node == NULL)
            return ValueIterator(NULL);
        else
            return ValueIterator(node);
    }
	

//     /**
//      * Returns a string representation of the tree state of this, i.e., the
//      * concrete state. (The version of toString commented out below returns a
//      * representation of the abstract state of this->
//      */
//     public: String toString() {
//         StringBuffer buf = new StringBuffer();
//         buf.append("<root>"); //$NON-NLS-1$
//         toStringHelper(root, buf, 1);
//         return buf.toString();
//     }

// private:
//     /**
//      * Prints a description of the substree starting with start to buf. The
//      * printing starts with the given indent level. (internal)
//      */
// 	void toStringHelper(TrieNode start, StringBuffer buf, int indent) {
//         // Print value of node.
//         if (start.getValue() != NULL) {
//             buf.append(" -> "); //$NON-NLS-1$
//             buf.append(start.getValue().toString());
//         }
//         buf.append("\n"); //$NON-NLS-1$
//         // For each child...
//         for (Iterator iter = start.labelsForward(); iter.hasNext();) {
//             // Indent child appropriately.
//             for (int i = 0; i < indent; i++)
//                 buf.append(" "); //$NON-NLS-1$
//             // Print edge.
//             String label = (String) iter.next();
//             buf.append(label);
//             // Recurse to print value.
//             TrieNode child = start.get(label.charAt(0)).getChild();
//             toStringHelper(child, buf, indent + 1);
//         }
//     }
};

CL_NS_END2
#endif
