/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "MorphAnalyzer.h"
#include "PatternConstants.h"
#include "../utils/DictionaryUtil.h"
#include "../utils/EomiUtil.h"
#include "../utils/IrregularUtil.h"
#include "../utils/MorphUtil.h"
#include "../utils/NounUtil.h"
#include "../utils/SyllableUtil.h"
#include "../utils/VerbUtil.h"
//#include <iostream>

CL_NS_DEF2(analysis, kr)

	
void MorphAnalyzer::analyze(const String& input,
							/*out*/ AnalysisOutputList& results) throw(MorphException) {
	if(input.endsWith(_T(".")))
		return analyze(input.substring(0,input.length()-1), POS_END, results);
	
	return analyze(input, POS_MID, results);
}
	
void MorphAnalyzer::analyze(const String& input, int pos,
							/*out*/ AnalysisOutputList& results) throw(MorphException) {		
	AnalysisOutputList candidates;
		
	bool josaFlag = true;
	bool eomiFlag = true;
				
	int strlen = input.length();
		
	//bool isVerbOnly = MorphUtil::hasVerbOnly(input);
	bool isVerbOnly = false;
	analysisWithEomi(input, _T(""), candidates);
		
	for(int i=strlen-1;i>0;i--) {
			
		String stem = input.substring(0,i);
		String eomi = input.substring(i);

		const String& feature = SyllableUtil::getFeature(eomi.charAt(0));		
		if(!isVerbOnly && josaFlag && feature[SyllableUtil::IDX_JOSA1]==_T('1')) {				
			analysisWithJosa(stem,eomi,candidates);
		}
			
		if(eomiFlag) {			
			analysisWithEomi(stem,eomi,candidates);
		}			
			
		if(josaFlag&&feature[SyllableUtil::IDX_JOSA2]==_T('0')) josaFlag = false;
		if(eomiFlag&&feature[SyllableUtil::IDX_EOMI2]==_T('0')) eomiFlag = false;
			
		if(!josaFlag && !eomiFlag) break;
	}
		
	sort(candidates.begin(), candidates.end(), AnalysisOutputComparator());

	if(!isVerbOnly) addSingleWord(input,candidates);

	bool changed = false;
	AnalysisOutputList::iterator iter;
	for (iter = candidates.begin(); iter != candidates.end(); iter++) {
		if(iter->getPatn()<=PatternConstants::PTN_NJ &&
		   iter->getScore()!=AnalysisOutput::SCORE_CORRECT &&
		   iter->getStem().length()>2) {
			NounUtil::confirmCNoun(*iter);
			changed=true;
		}
	}
	if(changed) sort(candidates.begin(), candidates.end(), AnalysisOutputComparator());
//		if(candidates.size()>0&&
//		   candidates.get(0).getScore()!=AnalysisOutput::SCORE_CORRECT) { // 분석에 실패하였다면 복합명사인지 조사한다.
//			 analysisCNoun(candidates);
//		}		
		
	bool hasCorrect = false;
	bool hasCorrectNoun = false;

	OutputMap stems;
	AnalysisOutput *noun = NULL;
	for (iter = candidates.begin(); iter != candidates.end(); iter++) {
//		std::wcout << _T("candidates=>") << iter->toString() << _T("<") << iter->getScore() << _T("> ") << iter->getPatn() << std::endl;
		if(iter->getScore()==AnalysisOutput::SCORE_CORRECT && iter->getPos()!=PatternConstants::POS_NOUN ) {
			addResults(*iter,results,stems);
			hasCorrect = true;				
		}
		else if(iter->getPos()==PatternConstants::POS_NOUN&&iter->getScore()==AnalysisOutput::SCORE_CORRECT) {
//			TCHAR *key = iter->getStem();
//			if(iter->getCNounList().size()>0) key = iter->getCNounList().at(0);
//			if(nouns.get(key)==null) {
//				nouns.put(key, *iter);
//				results.add(*iter);
//			}				
			if(iter->getPos()==PatternConstants::POS_NOUN) {
				addResults(*iter,results,stems);
			}
			/* FIXME: 코드가 이상함, 밑에 2조건은 절대 실행이 안됨 */
			else if (noun==NULL) {
				addResults(*iter,results,stems);
				noun = &*iter;
			}
			else if (iter->getPatn()==PatternConstants::PTN_N || (iter->getPatn() > noun->getPatn()) ||
					(iter->getPatn()==noun->getPatn() &&
					 !iter->getJosa().empty() &&
					 !noun->getJosa().empty() &&
					 iter->getJosa().length() > noun->getJosa().length()) )
			{
				AnalysisOutputList::iterator resultsIter = results.begin();
				for (; resultsIter != results.end(); resultsIter++) {
					if(&*iter == noun)
						break;
				}
				results.erase(resultsIter);
				addResults(*iter,results,stems);
				noun = &*iter;
			}
			hasCorrectNoun=true;				
		}
		else if(iter->getPos()==PatternConstants::POS_NOUN&&!hasCorrect&&!hasCorrectNoun) {
			addResults(*iter,results,stems);
		}
	}				

	return;
}

// CHECK: stem value of AnalysisOutput in results is unique?
void MorphAnalyzer::addResults(AnalysisOutput& o, AnalysisOutputList& results, OutputMap& stems) {
	OutputMap::iterator old = stems.find(o.getStem());
	if(old == stems.end()) {
		results.push_back(o);
		stems[o.getStem()] = o;
	}
	else if(old->second.getPatn() < o.getPatn()) {
		for (AnalysisOutputList::iterator iter = results.begin();
			 iter != results.end(); iter++) {
			if (iter->getStem().equals(old->second.getStem())) {
				*iter = o;
				stems[o.getStem()] = o;
				break;
			}
		}
	}
}
	
void MorphAnalyzer::addSingleWord(const String& word, AnalysisOutputList& candidates)
	throw(MorphException) {
	
	AnalysisOutput output(word, _T(""), _T(""), PatternConstants::PTN_N);
	output.setPos(PatternConstants::POS_NOUN);

	DictionaryUtil::EntryPosition entry;
	if(!(entry=DictionaryUtil::findWord(word)).isNull()) {

		if(entry->getFeature(WordEntry::IDX_BUSA)==_T('1')) {
			AnalysisOutput busa(word, _T(""), _T(""), PatternConstants::PTN_AID);
			busa.setPos(PatternConstants::POS_ETC);
				
			busa.setScore(AnalysisOutput::SCORE_CORRECT);
			candidates.insert(candidates.begin(), busa);		
		}else if(entry->getFeature(WordEntry::IDX_NOUN)==_T('1')) {
			output.setScore(AnalysisOutput::SCORE_CORRECT);
			candidates.insert(candidates.begin(), output);
		}else if(entry->getFeature(WordEntry::IDX_NOUN)==_T('2')) {
			candidates.insert(candidates.begin(), output);
		}
			
		if(entry->getFeature(WordEntry::IDX_VERB)!=_T('1')) return;
	} 

	if(candidates.size()==0||candidates.at(0).getScore()!=AnalysisOutput::SCORE_CORRECT
	   ||!NounUtil::endsWith2Josa(word)) candidates.push_back(output);			
}
	
void MorphAnalyzer::analysisWithJosa(const String& stem, const String& end,
									 AnalysisOutputList& candidates) throw(MorphException) {
	
	if(stem.length()==0) return;	
	if(!DictionaryUtil::existJosa(end)) return;
		
	AnalysisOutput output(stem, end, _T(""), PatternConstants::PTN_NJ);
	output.setPos(PatternConstants::POS_NOUN);
		
	bool success = false;	
	success = NounUtil::analysisMJ(output, candidates);

	DictionaryUtil::EntryPosition entry = DictionaryUtil::findNoun(stem);
	if(!entry.isNull()) {
		output.setScore(AnalysisOutput::SCORE_CORRECT);
//	} else {
//			NounUtil::confirmDNoun(output);
	}
		
	candidates.push_back(output);

}

void MorphAnalyzer::analysisWithEomi(const String& stem, const String& end,
									 AnalysisOutputList& candidates) throw(MorphException) {

	String morphs[2];
	EomiUtil::splitEomi(stem, end, morphs);
	if(morphs[0].empty()) return; // 어미가 사전에 등록되어 있지 않다면....

	String pomis[2];
	EomiUtil::splitPomi(morphs[0], pomis);

	AnalysisOutput o(pomis[0], _T(""), morphs[1], PatternConstants::PTN_VM);
	o.setPomi(pomis[1]);
			
	DictionaryUtil::EntryPosition entry = DictionaryUtil::findVerb(o.getStem());
	if(!entry.isNull()) {							
		AnalysisOutput output = o;
		output.setScore(AnalysisOutput::SCORE_CORRECT);
		MorphUtil::buildPtnVM(output, candidates);
		return;
	}

	String irrs[2];
	if (IrregularUtil::restoreIrregularVerb(o.getStem(),
											o.getPomi().empty() ? o.getEomi() : o.getPomi(),
											irrs) ) {  // 불규칙동사인 경우
		AnalysisOutput output = o;
		output.setStem(irrs[0]);
		if(output.getPomi().empty())
			output.setEomi(irrs[1]);
		else
			output.setPomi(irrs[1]);
		
//		entry = DictionaryUtil::findVerb(output.getStem());
//		if(entry!=null && VerbUtil::constraintVerb(o.getStem(), o.getPomi()==null?o.getEomi():o.getPomi())) { // 4. 돕다 (PTN_VM)
			output.setScore(AnalysisOutput::SCORE_CORRECT);
			MorphUtil::buildPtnVM(output, candidates);
//		}				
	}
			
	if(VerbUtil::analysisNSM(o, candidates)) return;
	
	if(VerbUtil::analysisNSMXM(o, candidates)) return;
	
	// [체언 + '에서/에서부터' + _T('이') +  어미]
	if(VerbUtil::analysisNJCM(o, candidates)) return;			
	
	if(VerbUtil::analysisVMCM(o, candidates)) return;	
	
	VerbUtil::analysisVMXM(o, candidates);		
}		

void MorphAnalyzer::analysisCNoun(AnalysisOutputList& candidates) throw(MorphException) {		
	bool success = false;
	AnalysisOutputList::iterator iter = candidates.begin();
	for (; iter != candidates.end(); iter++) {
		if (iter->getPos() != PatternConstants::POS_NOUN) continue;
		if (iter->getScore() == AnalysisOutput::SCORE_CORRECT) 
			success=true;
		else if(!success)
			NounUtil::confirmCNoun(*iter);
	}
}

CL_NS_END2
