/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef _lucene_analysis_kr_morph_MorphAnalyzer
#define _lucene_analysis_kr_morph_MorphAnalyzer

#if defined(_LUCENE_PRAGMA_ONCE)
# pragma once
#endif

#include "CLucene/StdHeader.h"
#include "AnalysisOutput.h"
#include "MorphException.h"
#include <ext/hash_map>

CL_NS_DEF2(analysis, kr)

class MorphAnalyzer {
private:
	struct StringHash {
		// FIXME: overflow? (wchar_t may be 4bytes)
		size_t operator()(const String& _Key) const	{
			unsigned long __h = 0;
			for (int i = 0 ; i < _Key.length(); i++)
				__h = 5 * __h + _Key[i];
			return size_t(__h);
		}		
	};

	struct StringEqual {
		bool operator()(const String& s1, const String& s2) const {
			return (s1.compare(s2) == 0) ? true : false;
		}
	};
	typedef __gnu_cxx::hash_map<String, AnalysisOutput, StringHash, StringEqual> OutputMap;
	
public:
	/**
	 * starting word of sentence.
	 */
	static const int POS_START = 1;
	
	/**
	 * middle word of sentence
	 */
	static const int POS_MID = 2;
	
	/**
	 * ending word of sentence.
	 */
	static const int POS_END = 3;
	
	
	void analyze(const String& input, AnalysisOutputList& result) throw(MorphException);
	
	/**
	 * 
	 * @param input
	 * @param pos
	 * @return
	 * @throws MorphException
	 */
	void analyze(const String& input, int pos, AnalysisOutputList& results) throw(MorphException);

private:
	void addResults(AnalysisOutput& o, AnalysisOutputList& results, OutputMap& stems);
	
	void addSingleWord(const String& word, AnalysisOutputList& candidates) throw(MorphException);
	
	/**
	 * 체언 + 조사 (PTN_NJ)
	 * 체언 + 용언화접미사 + '음/기' + 조사 (PTN_NSMJ
	 * 용언 + '음/기' + 조사 (PTN_VMJ)
	 * 용언 + '아/어' + 보조용언 + '음/기' + 조사(PTN_VMXMJ)
	 * 
	 * @param stem
	 * @param end
	 * @param candidates
	 * @throws MorphException
	 */
	void analysisWithJosa(const String& stem, const String& end,
						  AnalysisOutputList& candidates) throw(MorphException);
		
	/**
	 * 
	 *  1. 사랑받다 : 체언 + 용언화접미사 + 어미 (PTN_NSM) <br>
	 *  2. 사랑받아보다 : 체언 + 용언화접미사 + '아/어' + 보조용언 + 어미 (PTN_NSMXM) <br>
	 *  3. 학교에서이다 : 체언 + '에서/부터/에서부터' + '이' + 어미 (PTN_NJCM) <br>
	 *  4. 돕다 : 용언 + 어미 (PTN_VM) <br>
	 *  5. 도움이다 : 용언 + '음/기' + '이' + 어미 (PTN_VMCM) <br>
	 *  6. 도와주다 : 용언 + '아/어' + 보조용언 + 어미 (PTN_VMXM) <br>
	 *  
	 * @param stem
	 * @param end
	 * @param candidates
	 * @throws CloneNotSupportedException 
	 */
	void analysisWithEomi(const String& stem, const String& end,
						  AnalysisOutputList& candidates) throw(MorphException);

public:
	void analysisCNoun(AnalysisOutputList& candidates) throw(MorphException);

};

CL_NS_END2
#endif
