/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "../config.h"
#include "DictionaryUtil.h"

CL_NS_DEF2(analysis, kr)

DictionaryUtil::Dictionary* DictionaryUtil::dictionary = NULL;

DictionaryUtil::Dictionary::Dictionary(): trie(true) {
	StringList strList;
	StringList compounds;
	
	StringList josaList;
	StringList eomiList;
	StringList prefixList;
	StringList suffixList;
	
	try {	
		FileUtil::readLines(FILE_DICTIONARY, "UTF-8", strList);
		FileUtil::readLines(FILE_EXTENSION,  "UTF-8", strList);
		FileUtil::readLines(FILE_COMPOUNDS,  "UTF-8", compounds);
		
		FileUtil::readLines(FILE_JOSA,   "UTF-8", josaList);
		FileUtil::readLines(FILE_EOMI,   "UTF-8", eomiList);
		FileUtil::readLines(FILE_PREFIX, "UTF-8", prefixList);
		FileUtil::readLines(FILE_SUFFIX, "UTF-8", suffixList);
	}
	catch (const IOException &e) {			
		throw MorphException(e.getMessage());
	}
	
	if (strList.size() == 0)
		throw MorphException("dictionary is null");

	StringList::iterator iter = ++strList.begin();
	for(; iter != strList.end(); iter++) {
		StringList infos;
		iter->split(_T(','), infos);
		if (infos.size() != 2) continue;
		String word = infos[0].trim();
		trie.add(word, new WordEntry(word, infos[1].trim()) );
	}

	for(iter = ++compounds.begin(); iter != compounds.end(); iter++) {
		StringList infos;
		iter->split(_T(':'), infos);
		if (infos.size() != 2) continue;
		String word = infos[0].trim();		
		StringList elements;
		CompoundEntryList compoundEntries;
		infos[1].split(_T(','), elements);
		compoundArrayToList(infos[1], elements, compoundEntries);
		trie.add(word, new WordEntry(word, _T("20000X"), compoundEntries));
	}

	for (iter = ++josaList.begin(); iter != josaList.end(); iter++) {
		josas[iter->trim()] = *iter;
	}
	for (iter = ++eomiList.begin(); iter != eomiList.end(); iter++) {
		eomis[iter->trim()] = *iter;
	}
	for (iter = ++prefixList.begin(); iter != prefixList.end(); iter++) {
		prefixs[iter->trim()] = *iter;
	}	
	for(iter = ++suffixList.begin(); iter != suffixList.end(); iter++) {
		suffixs[iter->trim()] = *iter;
	}	
}


/**
 * 사전을 로드한다.
 */
void DictionaryUtil::loadDictionary() throw(MorphException) {
	DEFINE_MUTEX(THIS_LOCK)
	SCOPED_LOCK_MUTEX(THIS_LOCK)

	if (DictionaryUtil::dictionary != NULL)
		return;
	
	static Dictionary dictionary;
	DictionaryUtil::dictionary = &dictionary;
}


DictionaryUtil::EntryIterator
DictionaryUtil::findWithPrefix(const String& prefix) throw(MorphException) {
	if (dictionary==NULL) loadDictionary();
	return dictionary->trie.getPrefixedBy(prefix);
}

DictionaryUtil::EntryPosition
DictionaryUtil::findWord(const String& key) throw(MorphException) {
	if (dictionary==NULL) loadDictionary();
	return dictionary->trie.find(key);
}
	
DictionaryUtil::EntryPosition
DictionaryUtil::findNoun(const String& key) throw(MorphException) {
	EntryPosition entry = findWord(key);
	if(entry.isNull()) return EntryPosition(NULL);
		
	if(entry->getFeature(WordEntry::IDX_NOUN)=='1') return entry;
	return EntryPosition(NULL);;
}

// CHECK: WordEntry::IDX_NOUN == '2' ???
DictionaryUtil::
EntryPosition DictionaryUtil::findCNoun(const String& key) throw(MorphException) {	
	EntryPosition entry = findWord(key);
	if(entry.isNull()) return EntryPosition(NULL);

	if(entry->getFeature(WordEntry::IDX_NOUN)=='1' || entry->getFeature(WordEntry::IDX_NOUN)=='2') return entry;
	return EntryPosition(NULL);
}

DictionaryUtil::EntryPosition
DictionaryUtil::findVerb(const String& key) throw(MorphException) {
		
	EntryPosition entry = findWord(key);	
	if(entry.isNull()) return EntryPosition(NULL);

	if(entry->getFeature(WordEntry::IDX_VERB)=='1') {
		return entry;
	}
	return EntryPosition(NULL);
}
	
DictionaryUtil::EntryPosition
DictionaryUtil::findAdverb(const String& key) throw(MorphException) {
	EntryPosition entry = findWord(key);
	if(entry.isNull()) return EntryPosition(NULL);

	if(entry->getFeature(WordEntry::IDX_BUSA)=='1') return entry;
	return EntryPosition(NULL);
}
	
DictionaryUtil::EntryPosition
DictionaryUtil::findIrrVerb(const String& key, TCHAR irrType) throw(MorphException) {
	EntryPosition entry = findWord(key);
	if(entry.isNull()) return EntryPosition(NULL);

	if(entry->getFeature(WordEntry::IDX_VERB)=='1'&&
	   entry->getFeature(WordEntry::IDX_REGURA)==irrType) return entry;
	return EntryPosition(NULL);
}
	
DictionaryUtil::EntryPosition
DictionaryUtil::findBeVerb(const String& key) throw(MorphException) {
	EntryPosition entry = findWord(key);
	if(entry.isNull()) return EntryPosition(NULL);
		
	if(entry->getFeature(WordEntry::IDX_BEV)=='1') return entry;
	return EntryPosition(NULL);
}
	
DictionaryUtil::EntryPosition
DictionaryUtil::findDoVerb(const String& key) throw(MorphException) {
	EntryPosition entry = findWord(key);
	if(entry.isNull()) return EntryPosition(NULL);
		
	if(entry->getFeature(WordEntry::IDX_DOV)=='1') return entry;
	return EntryPosition(NULL);
}
	
bool DictionaryUtil::existJosa(const String& str) throw(MorphException) {
	if (dictionary==NULL) loadDictionary();
	if (dictionary->josas.find(str) == dictionary->josas.end()) return false;
	else return true;
}
	
bool DictionaryUtil::existEomi(const String& str) throw(MorphException) {
	if (dictionary==NULL) loadDictionary();
	if (dictionary->eomis.find(str) == dictionary->eomis.end()) return false;
	else return true;
}
	
bool DictionaryUtil::existPrefix(const String& str) throw(MorphException) {
	if (dictionary==NULL) loadDictionary();
	if (dictionary->prefixs.find(str) == dictionary->prefixs.end()) return false;
	else return true;
}
	
bool DictionaryUtil::existSuffix(const String& str) throw(MorphException) {
	if (dictionary==NULL) loadDictionary();
	if (dictionary->suffixs.find(str) != dictionary->suffixs.end()) return true;		
	return false;
}
	
/**
 * ㄴ,ㄹ,ㅁ,ㅂ과 eomi 가 결합하여 어미가 될 수 있는지 점검한다.
 * @param s
 * @param end
 * @return
 */
String DictionaryUtil::combineAndEomiCheck(TCHAR s, const String& eomi) throw(MorphException) {
	String _eomi;	
	if (s == _T('ㄴ'))		 _eomi = _T("은") + eomi;
	else if (s == _T('ㄹ')) _eomi = _T("을") + eomi;
	else if (s == _T('ㅁ')) _eomi = _T("음") + eomi;
	else if (s == _T('ㅂ')) _eomi = _T("습") + eomi;
	else _eomi = s + eomi;

	if (existEomi(_eomi))
		return _eomi;		
	else
		return _T("");		
}
	

void DictionaryUtil::compoundArrayToList(const String& source, const StringList& arr,
						 /*out*/ CompoundEntryList& list) {
	StringList::const_iterator iter = arr.begin();
	for (; iter != arr.end(); iter++) {
		list.push_back(CompoundEntry(*iter, source.find(*iter)) );
	}
}

CL_NS_END2
