/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef _lucene_analysis_kr_utils_EomiUtil
#define _lucene_analysis_kr_utils_EomiUtil

#if defined(_LUCENE_PRAGMA_ONCE)
# pragma once
#endif

#include "CLucene/StdHeader.h"
#include "../morph/AnalysisOutput.h"
#include "../morph/PatternConstants.h"
#include "../morph/MorphException.h"
#include "SyllableUtil.h"
#include "String.h"

CL_NS_DEF2(analysis, kr)

class EomiUtil {
public:	
	static const TCHAR* RESULT_FAIL;
	static const TCHAR* RESULT_SUCCESS;
	static const TCHAR* verbSuffix[];
	
	/**
	 * 가장 길이가 긴 어미를 분리한다.
	 * @param term
	 * @return
	 * @throws MorphException
	 */
	static void longestEomi(const String& term, String result[]) throw(MorphException);
	
	/**
	 * 선어말어미를 분석한다.
	 * @param stem
	 * @return
	 */
	static void splitPomi(const String& stem, String results[]) throw(MorphException);
	
	/**
	 * 불규칙 용언의 원형을 구한다.
	 * @param output
	 * @return
	 * @throws MorphException
	 */
	static void irregular(AnalysisOutput& output, /*out*/ AnalysisOutputList& results);

private:
	/**
	 * 어간만 변하는 경우
	 * @param results
	 * @param stem
	 * @param ending
	 */
	static void irregularStem(StringListList& results, const String& stem, const String& ending);
	
	/**
	 * 어미만 변하는 경우
	 * @param results
	 * @param stem
	 * @param ending
	 */
	static void irregularEnding(StringListList& results, const String& stem, const String& ending);
	
	/**
	 * 어간과 어미가 모두 변하는 경우
	 * @param results
	 * @param stem
	 * @param ending
	 */
	static void irregularAO(StringListList& results, const String& stem, const String& ending);
	
	/**
	 * 시제 선어미말을 만들어서 반환한다.
	 * @param preword  '아' 또는 '어'
	 * @param endword  어미[선어미말을 포함]
	 * @return '았' 또는 '었'을 만들어서 반환한다.
	 */
	static String makeTesnseEomi(const String& preword, const String& endword);	
	
	/**
	 * 용언화 접미사가 있는지 조사한다.
	 * @param result
	 * @return
	 */
    static bool hasVerbSuffix(const String& stem) {
	
		for (int i = 0; verbSuffix[i] != NULL; i++) {
			if(stem.endsWith(String(verbSuffix[i])) ) return true;
		}
		
		return false;
   }

	/**
	 * 용언화 접미사가 있는지 조사한다.
	 * @param result
	 * @return
	 */
	inline static bool isVerbSuffix(String stem) {
	
		for (int i = 0; verbSuffix[i] != NULL; i++) {
			if(stem.equals(verbSuffix[i])) return true;
		}
		
		return false;
  }
  
	static void setPomiResult(String results[], const String& stem, const String& pomi ) {
		results[0] = stem;
		results[1] = pomi;
	}	

public:
	/**
	 * 
	 * @param ch
	 * @return
	 */
	static bool IsNLMBSyl(TCHAR ech, TCHAR lch) throw(MorphException) {
	
		const String& features = SyllableUtil::getFeature(ech);

		if (lch == _T('ㄴ')) {
			return (features[SyllableUtil::IDX_YNPNA]==_T('1'));
		}
		else if (lch == _T('ㄹ')) {
			return (features[SyllableUtil::IDX_YNPLA]==_T('1'));
		}
		else if (lch ==  _T('ㅁ')) {
			return (features[SyllableUtil::IDX_YNPMA]==_T('1'));
		}
		else if (lch ==  _T('ㅂ')) {
			return (features[SyllableUtil::IDX_YNPBA]==_T('1'));
		}

		return false;
	}

	static bool endsWithEEomi(const String& stem);
		
	/**
	 * 어미를 분리한다.
	 * 
	 * 1. 규칙용언과 어간만 바뀌는 불규칙 용언
	 * 2. 어미가 종성 'ㄴ/ㄹ/ㅁ/ㅂ'으로 시작되는 어절
	 * 3. '여/거라/너라'의 불규칙 어절
	 * 4. 어미 '아/어'가 탈락되는 어절
	 * 5. '아/어'의 변이체 분리
	 * 
	 * @param stem
	 * @param end
	 * @return
	 * @throws MorphException
	 */
	static void splitEomi(const String& stem, const String& end,
						  /*out*/ String strs[]) throw(MorphException);	
};

CL_NS_END2
#endif
