/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef _lucene_analysis_kr_utils_FileUtil
#define _lucene_analysis_kr_utils_FileUtil

#if defined(_LUCENE_PRAGMA_ONCE)
# pragma once
#endif

#include "CLucene/StdHeader.h"
#include "CLucene/util/Reader.h"
#include "String.h"

CL_NS_USE(util)

CL_NS_DEF2(analysis, kr)

class IOException {
	char *message;
public:
	IOException(): message(NULL) {}	
	IOException(const char *msg) { message = strdup(msg); }
	~IOException() { free(message); }	
	const char* getMessage() const { return message; }
};

class FileUtil {
public:	
    /**
     * Reads the contents of a file line by line to a List of Strings.
     * The file is always closed.
     *
     * @param file  the file to read, must not be <code>null</code>
     * @param encoding  the encoding to use, <code>null</code> means platform default
     * @return the list of Strings representing each line in the file, never <code>null</code>
     * @throws Exception 
     * @throws java.io.UnsupportedEncodingException if the encoding is not supported by the VM
     * @since Commons IO 1.1
     */
	// CHECK: Is it inefficient?
    static void readLines(const char *fName, const char *encoding,
						  StringList &result) throw (IOException) {
		const TCHAR *c;
		String buf;
		int nread;
		FileReader in(fName, encoding);

		// A line is considered to be terminated by any one of a line feed ('\n'),
		// a carriage return ('\r'), or a carriage return followed immediately by a linefeed
		while ((nread = in.read(c, 1, 1)) >= 0) {
			if (nread == 0) continue;
			
			if (*c == '\r' || *c == '\n') {
				if (buf.length() == 0) continue;
				result.push_back(buf);
				buf.clear();
			}
			else {
				buf.append(1, *c);
			}
		}
		if (nread == -2)
			throw IOException(std::string("Could not read file: ").append(fName).c_str());
		
		if (buf.length() > 0)
			result.push_back(buf);
	}
};

CL_NS_END2
#endif
