/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef _lucene_analysis_kr_utils_IrregularUtil
#define _lucene_analysis_kr_utils_IrregularUtil

#if defined(_LUCENE_PRAGMA_ONCE)
# pragma once
#endif

#include "CLucene/StdHeader.h"
#include "String.h"
#include "../morph/MorphException.h"


CL_NS_DEF2(analysis, kr)

/**
 * 
 * 동사의 불규칙 변형을 처리하는 Utility Class
 * 
 * @author S.M.Lee
 *
 */
class IrregularUtil {
public:	
	// ㅂ 불규칙
	static const TCHAR IRR_TYPE_BIUP = _T('B');
	
	// ㅎ 불규칙
	static const TCHAR IRR_TYPE_HIOOT = _T('H');
	
	// ㄹ 불규칙
	static const TCHAR IRR_TYPE_LIUL = _T('X');
	
	// ㄹ 불규칙
	static const TCHAR IRR_TYPE_LOO = _T('L');

	// ㅅ 불규칙
	static const TCHAR IRR_TYPE_SIUT = _T('S');
	
	// ㄷ 불규칙
	static const TCHAR IRR_TYPE_DI = _T('D');
	
	// 러 불규칙
	static const TCHAR IRR_TYPE_RU = _T('R');
	
	// 으 탈락
	static const TCHAR IRR_TYPE_UI = _T('X');	
	
	// 규칙형
	static const TCHAR IRR_TYPE_REGULAR = _T('X');
	
	static bool restoreIrregularVerb(const String& start, const String& end,
									 /*out*/ String results[]) throw(MorphException);


private:
	/**
	 * ㅂ 불규칙 원형을 복원한다. (돕다, 곱다)
	 * @param start
	 * @param end
	 * @return
	 * @throws MorphException
	 */
	static bool restoreBIrregular(const String& start, const String& end,
								  /*out*/ String results[]) throw(MorphException);

	
	/**
	 * ㄷ 불규칙 원형을 복원한다. (깨닫다, 묻다)
	 * @param start
	 * @param end
	 * @return
	 * @throws MorphException
	 */
	static bool restoreDIrregular(const String& start, const String& end,
								  /*out*/ String results[]) throw(MorphException);

	
	/**
	 * ㅅ 불규칙 원형을 복원한다. (긋다--그어)
	 * @param start
	 * @param end
	 * @return
	 * @throws MorphException
	 */
	static bool restoreSIrregular(const String& start, const String& end,
								  /*out*/ String results[]) throw(MorphException);
		
	/**
	 * 르 불규칙 원형을 복원한다. (흐르다-->흘러)
	 * @param start
	 * @param end
	 * @return
	 * @throws MorphException
	 */
	static bool restoreLIrregular(const String& start, const String& end,
								  /*out*/ String results[]) throw(MorphException);
	
	/**
	 * ㄹ불규칙 원형을 복원한다. (길다-->긴, 알다-->안)
	 * 어간의 끝소리인 ‘ㄹ’이 ‘ㄴ’, ‘ㄹ’, ‘ㅂ’, ‘오’, ‘시’ 앞에서 탈락하는 활용의 형식
	 * @param start
	 * @param end
	 * @return
	 * @throws MorphException
	 */
	static bool restoreELIrregular(const String& start, const String& end,
								   /*out*/ String results[]) throw(MorphException);
	
	/**
	 * 러 불규칙 원형을 복원한다. (이르다->이르러, 푸르다->푸르러)
	 * @param start
	 * @param end
	 * @return
	 * @throws MorphException
	 */
	static bool restoreRUIrregular(const String& start, const String& end,
								   /*out*/ String results[]) throw(MorphException);
	
	/**
	 * ㅎ 탈락 원형을 복원한다. (까맣다-->까만,까매서)
	 * @param start
	 * @param end
	 * @return
	 * @throws MorphException
	 */
	static bool restoreHIrregular(const String& start, const String& end,
								  /*out*/ String results[]) throw(MorphException);
	/**
	 * 으 탈락 원형을 복원한다. (뜨다->더, 크다-커)
	 * @param start
	 * @param end
	 * @return
	 * @throws MorphException
	 */
	static bool restoreUIrregular(const String& start, const String& end,
								  /*out*/ String results[]) throw(MorphException);	
};


CL_NS_END2
#endif
