/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "MorphUtil.h"
#include "SyllableUtil.h"
#include "DictionaryUtil.h"
#include "IrregularUtil.h"
#include "../morph/PatternConstants.h"

CL_NS_USE(analysis)
CL_NS_USE(util)

CL_NS_DEF2(analysis, kr)

const TCHAR MorphUtil::CHOSEONG[] = {
	_T('ㄱ'),_T('ㄲ'),_T('ㄴ'),_T('ㄷ'),_T('ㄸ'),_T('ㄹ'),_T('ㅁ'),_T('ㅂ'),_T('ㅃ'),_T('ㅅ'),
	_T('ㅆ'),_T('ㅇ'),_T('ㅈ'),_T('ㅉ'),_T('ㅊ'),_T('ㅋ'),_T('ㅌ'),_T('ㅍ'),_T('ㅎ')
};

const TCHAR MorphUtil::JUNGSEONG[] = {
	_T('ㅏ'),_T('ㅐ'),_T('ㅑ'),_T('ㅒ'),_T('ㅓ'),_T('ㅔ'),_T('ㅕ'),_T('ㅖ'),_T('ㅗ'),_T('ㅘ'),
	_T('ㅙ'),_T('ㅚ'),_T('ㅛ'),_T('ㅜ'),_T('ㅝ'),_T('ㅞ'),_T('ㅟ'),_T('ㅠ'),_T('ㅡ'),_T('ㅢ'),
	_T('ㅣ')
};
	
const TCHAR MorphUtil::JONGSEONG[] = {
	_T('\0'), _T('ㄱ'),_T('ㄲ'),_T('ㄳ'),_T('ㄴ'),_T('ㄵ'),_T('ㄶ'),_T('ㄷ'),_T('ㄹ'),_T('ㄺ'),
	_T('ㄻ'),_T('ㄼ'),_T('ㄽ'),_T('ㄾ'),_T('ㄿ'),_T('ㅀ'),_T('ㅁ'),_T('ㅂ'),_T('ㅄ'),_T('ㅅ'),
	_T('ㅆ'),_T('ㅇ'),_T('ㅈ'),_T('ㅊ'),_T('ㅋ'),_T('ㅌ'),_T('ㅍ'),_T('ㅎ')
};


/**
 * 형태소 유형 출력을 위한 문자열을 생성한다.
 * @param word
 * @param type
 * @return
 */
String MorphUtil::buildTypeString(const String& word, TCHAR type) {
	String sb(word);
	sb.append(_T("("));
	sb.append(1, type);
	sb.append(_T(")"));
		
	return sb;
}
	
	
void MorphUtil::buildPtnVM(AnalysisOutput& output, AnalysisOutputList& candidates) throw(MorphException) {
		
	const String& end = (output.getPomi().empty()) ? output.getEomi() :
													 output.getPomi();
		
	output.setPatn(PatternConstants::PTN_VM);
	output.setPos(PatternConstants::POS_VERB);
		
	if(output.getScore()==AnalysisOutput::SCORE_CORRECT) {
		candidates.push_back(output);
	}else {
		String irrs[2];
		if(IrregularUtil::restoreIrregularVerb(output.getStem(),end, irrs)) {
			output.setScore(AnalysisOutput::SCORE_CORRECT);
			output.setStem(irrs[0]);
			candidates.push_back(output);	
		}
	}
		
}
	
/**
 * 용언 + '음/기' + '이' + 어미, 체언 + '에서/부터/에서부터' + '이' + 어미
 * @param output
 * @param candidates
 * @throws MorphException
 */
void MorphUtil::buildPtnCM(AnalysisOutput& output, AnalysisOutputList& candidates) throw(MorphException) {
		
	TCHAR ch = output.getStem().charAt(output.getStem().length()-2);
	TCHAR jasos[3];
	int jasos_length = MorphUtil::decompose(ch, jasos);
	if(jasos_length == 3 || ch==_T('기')) {
		buildPtnVMCM(output,candidates);			
	} else {
			
	}
}


void MorphUtil::buildPtnVMCM(AnalysisOutput& output, AnalysisOutputList& candidates) throw(MorphException) {
	String stem = output.getStem();
	
	output.setPatn(PatternConstants::PTN_VMCM);
	output.setPos(PatternConstants::POS_VERB);
		
	TCHAR ch = stem.charAt(stem.length()-2);
	TCHAR jasos[3];
	int jasos_length = MorphUtil::decompose(ch, jasos);

	if (ch==_T('기')) {
		output.addElist(_T("기"));
		output.addElist(_T("이"));
		output.setStem(stem.substring(0,stem.length()-2));
			
		if (!DictionaryUtil::findVerb(output.getStem()).isNull())
			candidates.push_back(output);
	}
	else if(jasos[2]==_T('ㅁ')) {
		if(stem.length()>1) stem = stem.substring(0,stem.length()-2);
		stem += MorphUtil::makeChar(ch, 0);
		output.addElist(_T("ㅁ"));
		output.addElist(_T("이"));
		output.setStem(stem);

		if(!DictionaryUtil::findVerb(stem).isNull()) 
			candidates.push_back(output);
		else {
			String morphs[2];
			if (IrregularUtil::restoreIrregularVerb(stem, _T("ㅁ"), morphs)) {
				output.setScore(AnalysisOutput::SCORE_CORRECT);
				output.setStem(morphs[0]);
				candidates.push_back(output);
			}
		}
	}
}


bool MorphUtil::hasVerbOnly(const String& input) throw(MorphException) {
	for(int i = 0; i < input.length(); i++) {
		const String& feature = SyllableUtil::getFeature(input.charAt(i));
		if (feature[SyllableUtil::IDX_WDSURF]==_T('1') && input.length()>i) return true;
	}
	return false;
}
	
/**
 * 시제 선어미말을 만들어서 반환한다.
 * @param preword  '아' 또는 '어'
 * @param endword  어미[선어미말을 포함]
 * @return '았' 또는 '었'을 만들어서 반환한다.
 */
String MorphUtil::makeTesnseEomi(const String& preword, const String& endword) {

	if(preword.empty()) return endword;
	if(endword.empty()) return preword;

	if(endword.charAt(0) == _T('ㅆ')) {
		return preword.substring(0,preword.length()-1)
				.append(1, makeChar(preword.charAt(preword.length()-1),20))
				.append(endword.substring(1,endword.length()));
	}
	else if(endword.charAt(0) == _T('ㄴ')) {
		return preword.substring(0,preword.length()-1)
				.append(1, makeChar(preword.charAt(preword.length()-1),4))
				.append(endword.substring(1,endword.length()));
	}
	else if(endword.charAt(0) == _T('ㄹ')) {
		return preword.substring(0,preword.length()-1)
				.append(1, makeChar(preword.charAt(preword.length()-1),8))
				.append(endword.substring(1,endword.length()));	
	}
	else if(endword.charAt(0) == _T('ㅁ')) {
		return preword.substring(0,preword.length()-1)
				.append(1, makeChar(preword.charAt(preword.length()-1),16))
				.append(endword.substring(1,endword.length()));	
	}
	else if(endword.charAt(0) == _T('ㅂ')) {
		return preword.substring(0,preword.length()-1)
				.append(1, makeChar(preword.charAt(preword.length()-1),17))
				.append(endword.substring(1,endword.length()));
	}
	return String(preword).append(endword);
}

CL_NS_END2
