/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef _lucene_analysis_kr_utils_MorphUtil
#define _lucene_analysis_kr_utils_MorphUtil

#if defined(_LUCENE_PRAGMA_ONCE)
# pragma once
#endif

#include "CLucene/StdHeader.h"
#include "../morph/AnalysisOutput.h"
#include "../morph/MorphException.h"
#include "String.h"

CL_NS_USE(util)

CL_NS_DEF2(analysis, kr)


class MorphUtil {
private:
	static const TCHAR CHOSEONG[];
	static const TCHAR JUNGSEONG[];	
	static const TCHAR JONGSEONG[];	

	static const int JUNGSEONG_LENGTH = 21;
	static const int JONGSEONG_LENGTH = 28;
	
public:
	/**
	 * 한글 한글자를 초성/중성/종성의 배열로 만들어 반환한다.
	 * @param c
	 * @return
	 */
	static int decompose(TCHAR c, TCHAR *result) {
		result[0] = result[1] = result[2] = '\0';
		if(c>0xD7A3||c<0xAC00) { result[0] = c; return 1; }
		
		c -= 0xAC00;
		result[0] = CHOSEONG[c/(JUNGSEONG_LENGTH*JONGSEONG_LENGTH)];
		c = (TCHAR)(c % (JUNGSEONG_LENGTH*JONGSEONG_LENGTH));		
		result[1] = JUNGSEONG[c/JONGSEONG_LENGTH];		
		result[2] = JONGSEONG[c%JONGSEONG_LENGTH];
		
		return (result[2] != 0) ? 3 : 2;
	}
	
	static TCHAR compound(int first, int middle, int last) {		
		return (TCHAR)(0xAC00 +
					   first * (JUNGSEONG_LENGTH*JONGSEONG_LENGTH) +
					   middle * JONGSEONG_LENGTH +
					   last);
	}

	static TCHAR makeChar(TCHAR ch, int mdl, int last) {		
		ch -= 0xAC00;		
		int first = ch/(JUNGSEONG_LENGTH*JONGSEONG_LENGTH);		 
		return compound(first,mdl,last);
	}
	
	static TCHAR makeChar(TCHAR ch, int last) {
		ch -= 0xAC00;		
		int first = ch/(JUNGSEONG_LENGTH*JONGSEONG_LENGTH);	
		ch = (TCHAR)(ch % (JUNGSEONG_LENGTH*JONGSEONG_LENGTH));
		int middle = ch/JONGSEONG_LENGTH;
		
		return compound(first,middle,last);		
	}
	
	static TCHAR replaceJongsung(TCHAR dest, TCHAR source) {
		source -= 0xAC00;		
		int last = source % JONGSEONG_LENGTH;
			
		return makeChar(dest,last);	
	}

	/**
	 * 형태소 유형 출력을 위한 문자열을 생성한다.
	 * @param word
	 * @param type
	 * @return
	 */
	static String buildTypeString(const String& word, TCHAR type);
	
	
	static void buildPtnVM(AnalysisOutput& output, AnalysisOutputList& candidates) throw(MorphException);
	
	/**
	 * 용언 + '음/기' + '이' + 어미, 체언 + '에서/부터/에서부터' + '이' + 어미
	 * @param output
	 * @param candidates
	 * @throws MorphException
	 */
	static void buildPtnCM(AnalysisOutput& output, AnalysisOutputList& candidates) throw(MorphException);

private:
	static void buildPtnVMCM(AnalysisOutput& output, AnalysisOutputList& candidates) throw(MorphException);

public:
	static bool hasVerbOnly(const String& input) throw(MorphException);
	
	/**
	 * 시제 선어미말을 만들어서 반환한다.
	 * @param preword  '아' 또는 '어'
	 * @param endword  어미[선어미말을 포함]
	 * @return '았' 또는 '었'을 만들어서 반환한다.
	 */
	static String makeTesnseEomi(const String& preword, const String& endword);
};

CL_NS_END2
#endif
