/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "NounUtil.h"
#include "EomiUtil.h"
#include "SyllableUtil.h"
#include "DictionaryUtil.h"
#include "VerbUtil.h"
#include "IrregularUtil.h"
#include <memory>

CL_NS_DEF2(analysis, kr)

/**
 * 
 * 어간부가 음/기 로 끝나는 경우
 * 
 * @param o
 * @param candidates
 * @return
 * @throws MorphException
 */
bool NounUtil::analysisMJ(AnalysisOutput o, AnalysisOutputList& candidates) throw(MorphException) {

	int strlen = o.getStem().length();
		   
	if(strlen<2) return false;		   

	TCHAR chrs[3];
	int chrs_length = MorphUtil::decompose(o.getStem().charAt(strlen-1), chrs);
	bool success = false;

	if(o.getStem().charAt(strlen-1)!=_T('기')&&!(chrs_length==3&&chrs[2]==_T('ㅁ'))) return false;

	String start = o.getStem();
	String end = _T("");
	if(o.getStem().charAt(strlen-1)==_T('기')) {
		start = o.getStem().substring(0,strlen-1);
		end = _T("기");
	}else if(o.getStem().charAt(strlen-1)==_T('음')) {
		start = o.getStem().substring(0,strlen-1);
		end = _T("음");
	}

	String eomis[2];
	EomiUtil::splitEomi(start, end, eomis);
	if(eomis[0].empty()) return false;  // CHECK:
	String pomis[2];
	EomiUtil::splitPomi(eomis[0], pomis);
	o.setStem(pomis[0]);
	o.addElist(eomis[1]);		   
	o.setPomi(pomis[1]);
	   
	if(analysisVMJ(o,candidates)) return true;			   
	if(analysisVMXMJ(o,candidates)) return true;
	if(analysisNSMJ(o,candidates)) return true;	
		   		   
	if (!DictionaryUtil::findVerb(o.getStem()).isNull() ) {
		o.setPos(PatternConstants::POS_VERB);
		o.setPatn(PatternConstants::PTN_VMJ);
		o.setScore(AnalysisOutput::SCORE_CORRECT);
		candidates.push_back(o);
		return true;
	}
	   
	return false;
		  
}

/**
 * 용언 + '음/기' + 조사(PTN_VMXMJ)
 * @param o
 * @param candidates
 * @return
 * @throws MorphException
 */
bool NounUtil::analysisVMJ(AnalysisOutput o, AnalysisOutputList& candidates) throw(MorphException) {

	String irrs[2];	
	if (IrregularUtil::restoreIrregularVerb(o.getStem(), o.getElist().at(0), irrs)) {
		o.setStem(irrs[0]);
		o.setElist(irrs[1],0);
	}
			  
	if(!DictionaryUtil::findVerb(o.getStem()).isNull()) {
		o.setPatn(PatternConstants::PTN_VMJ);
		o.setPos(PatternConstants::POS_VERB);
		o.setScore(AnalysisOutput::SCORE_CORRECT);
		candidates.push_back(o);
		return true;
	}
			
	return false;
}
	  
/**
 * 용언 + '아/어' + 보조용언 + '음/기' + 조사(PTN_VMXMJ)
 * @param o
 * @param candidates
 * @return
 * @throws MorphException
 */
bool NounUtil::analysisVMXMJ(AnalysisOutput o, AnalysisOutputList& candidates) throw(MorphException) {

	int idxXVerb = VerbUtil::endsWithXVerb(o.getStem());

	if(idxXVerb!=-1) { // 2. 사랑받아보다
		String eogan = o.getStem().substring(0,idxXVerb);
		o.setXverb(o.getStem().substring(idxXVerb));
				
		String stomis[2];
		if(eogan.endsWith(_T("아"))||eogan.equals(_T("어")))
			EomiUtil::splitEomi(eogan.substring(0,eogan.length()-1),
								eogan.substring(eogan.length()-1),
								stomis);
		else
			EomiUtil::splitEomi(eogan, _T(""), stomis);
		
		if (stomis[0].empty()) return false;
				  
		String irrs[2];
		if(IrregularUtil::restoreIrregularVerb(stomis[0], stomis[1], irrs)) {
			o.setStem(irrs[0]);
			o.addElist(irrs[1]);
		} else {
			o.setStem(stomis[0]);
			o.addElist(stomis[1]);
		}
				  
		if(!DictionaryUtil::findVerb(o.getStem()).isNull()) {
			o.setPatn(PatternConstants::PTN_VMXMJ);
			o.setPos(PatternConstants::POS_VERB);
			o.setScore(AnalysisOutput::SCORE_CORRECT);
			candidates.push_back(o);
			return true;
		}
	}
			
	return false;
}
	  
/**
 * 체언 + 용언화접미사 + '음/기' + 조사 (PTN_NSMJ)
 * @param o
 * @param candidates
 * @return
 * @throws MorphException
 */
bool NounUtil::analysisNSMJ(AnalysisOutput o, AnalysisOutputList& candidates) throw(MorphException) {
		  
	int idxVbSfix = VerbUtil::endsWithVerbSuffix(o.getStem());
	if(idxVbSfix!=-1) { // 1. 사랑받다 (PTN_NSM)
		o.setVsfx(o.getStem().substring(idxVbSfix));
		o.setStem(o.getStem().substring(0,idxVbSfix));
		o.setPatn(PatternConstants::PTN_NSMJ);
		o.setPos(PatternConstants::POS_NOUN);
		if(!DictionaryUtil::findNoun(o.getStem()).isNull()) {
			o.setScore(AnalysisOutput::SCORE_CORRECT);
			candidates.push_back(o);
			return true;
//				}else {
//					NounUtil.confirmCNoun(o);
//					if(o.getScore()==AnalysisOutput::SCORE_CORRECT) {
//						candidates.add(o);
//						return true;
//					}
		}
	}
			
	return false;
}	  	   
	   
/**
 * 복합명사인지 조사하고, 복합명사이면 단위명사들을 찾는다.
 * 복합명사인지 여부는 단위명사가 모두 사전에 있는지 여부로 판단한다.
 * 단위명사는 2글자 이상 단어에서만 찾는다.
 * @param o
 * @return
 * @throws MorphException
 */
bool NounUtil::confirmCNoun(AnalysisOutput& o) throw(MorphException)  {
	typedef CL_NS(util)::CLVector<WordEntryList*, CL_NS(util)::Deletor::Void<WordEntryList> >
		WordEntryListList;
	
	if(o.getStem().length()<3) return false;
	if(o.getPatn()==PatternConstants::PTN_N &&
	   DictionaryUtil::existJosa(o.getStem().substring(o.getStem().length()-2))) return false;
		   
	CompoundEntryList results;
	WordEntryListList queue;
	String prefix = o.getStem().substring(0,1);
			
	int pos = 0;
	bool moreTwo =  false;
	while(pos < o.getStem().length()) {

		std::auto_ptr<WordEntryList> nList(new WordEntryList);
		if (! findNouns(o.getStem().substring(pos), queue.size(), o, *nList) ) 
			return false;
				
		if (pos==0 && DictionaryUtil::existPrefix(prefix))
			nList->push_back(WordEntry(prefix));
				
		if (nList->size()==0) {
			if(queue.size()==0) return false;
			WordEntryList *tmpList = queue.back();

			tmpList->pop_back();
			pos -= results.back().getWord().length();
			if(tmpList->size()==0) {				
				while(tmpList->size()==0) {				
					results.pop_back();
					queue.delete_back();
					if(queue.size()==0) return false;
							
					tmpList = queue.back();
					tmpList->pop_back();
					if(tmpList->size()==0) {
						pos -= results.back().getWord().length();
						continue;
					}
							
					pos -= results.back().getWord().length();			
					results.back() = CompoundEntry(tmpList->back().getWord(), pos);	
					pos += tmpList->back().getWord().length();
				}					
			}
			else {				
				results.back() = CompoundEntry(tmpList->back().getWord(), pos);
				pos += tmpList->back().getWord().length();
			}		

		} else {
			const WordEntry &noun = nList->back();
			queue.push_back(nList.release());
			results.push_back(CompoundEntry(noun.getWord(),pos));
			pos += noun.getWord().length();
			if(noun.getCompounds().size()>0) o.addCNoun(noun.getCompounds());
			if(noun.getWord().length()>1) moreTwo=true;
		}
	}

	if (results.size() > 1 && isDNoun(results.back().getWord()) ) {
		CompoundEntry dnoun = results.back();
		results.pop_back();
		o.setStem(o.getStem().substring(0, o.getStem().length()-dnoun.getWord().length()));
		o.setNsfx(dnoun.getWord());			
	}
			
	if(results.size()>1) o.addCNoun(results);	
			
	o.setScore(AnalysisOutput::SCORE_CORRECT);
	return true;
}


/**
 * 복합명사에서 단위명사를 분리해낸다.
 * 리스트의 가장 마지막에 위치한 단어가 최장단어이다.
 * @param str	복합명사
 * @param pos
 * @param o		분석결과
 * @return		단위명사 리스트
 * @throws MorphException
 */
bool NounUtil::findNouns(const String& str, int pos, AnalysisOutput& o,
			   /*out*/ WordEntryList& nList) throw(MorphException) {
	
	if (str.length()==2 &&
		DictionaryUtil::existSuffix(str.substring(0,1)) &&
		isDNoun(str.substring(1)) )
	{
		o.setStem(o.getStem().substring(0,o.getStem().length()-1));
		o.setNsfx(str.substring(1));
		nList.push_back(WordEntry(str.substring(0,1)));
		return true;
	}
	else if(str.length()==2 &&
			DictionaryUtil::existSuffix(str.substring(0,1)) &&
			DictionaryUtil::existJosa(str.substring(1))) {
		return false;
	}
			
	if(pos>=2 && DictionaryUtil::existJosa(str)) return false;
			
	if(str.length()==1 &&
	   (DictionaryUtil::existSuffix(str) || isDNoun(str)) ) {
		nList.push_back(WordEntry(str));
		return true;
	}

	for(int i=1;i<str.length();i++) {		
		String sub = str.substring(0,i+1);		
		if(!DictionaryUtil::findWithPrefix(sub).hasNext()) break;
		DictionaryUtil::EntryPosition entry = DictionaryUtil::findCNoun(sub);
		if(!entry.isNull()) {	
			nList.push_back(*entry);
		}
	}

	return true;			
}


/*
 * 마지막 음절이 명사형 접미사(등,상..)인지 조사한다.
 */
void NounUtil::confirmDNoun(AnalysisOutput& output) throw(MorphException) {

	int strlen = output.getStem().length();
	String d = output.getStem().substring(strlen-1);			
	if (!isDNoun(d)) return;

	String s = output.getStem().substring(0, strlen-1);
	if (!DictionaryUtil::findNoun(s).isNull())  {
		output.setStem(s);
		output.setNsfx(d);
		output.setScore(AnalysisOutput::SCORE_CORRECT);
	}
}
		
bool NounUtil::endsWith2Josa(const String& input) throw(MorphException) {

	bool josaFlag = true;
	for(int i=input.length()-2;i>0;i--) {
				
		String josa = input.substring(i);

		const String& feature =  SyllableUtil::getFeature(josa.charAt(0));		
		if(josaFlag&&DictionaryUtil::existJosa(josa)) return true;
	
				
		if(josaFlag&&feature[SyllableUtil::IDX_JOSA2]==_T('0')) josaFlag = false;				
		if(!josaFlag) break;
	}
			
	return false;
}


CL_NS_END2
