/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef _lucene_analysis_kr_utils_String
#define _lucene_analysis_kr_utils_String

#include "CLucene/StdHeader.h"
#include "CLucene/util/VoidList.h"
TCHAR cl_tcasefold(const TCHAR ch);


class String;
typedef CL_NS(util)::CLVector<String, CL_NS(util)::Deletor::NullVal<String> > StringList;
typedef CL_NS(util)::CLVector<StringList, CL_NS(util)::Deletor::NullVal<StringList> > StringListList;

/*
 * WARNING:
 * std::basic_string has no virtual functions (also destructor).
 * You should avoid allocating additional memory which need to be freed in the destructor.
 *
 */
  
class String : public std::basic_string<TCHAR> {	
public:
	explicit String() :
		std::basic_string<TCHAR>() {};

	String(const basic_string<TCHAR>& str) :
		std::basic_string<TCHAR>(str) {};

	String(const basic_string<TCHAR>& str, size_t pos, size_t n = npos) :
		std::basic_string<TCHAR>(str, pos, n) {};

	String(const TCHAR * s, size_t n) :
		std::basic_string<TCHAR>(s, n) {};
	
	String(const TCHAR * s) :
		std::basic_string<TCHAR>(s) {};
	
	String(size_t n, TCHAR c) :
		std::basic_string<TCHAR>(n, c) {};

	String(TCHAR c) :
		std::basic_string<TCHAR>(1, c) {};
	
	template<class InputIterator> String (InputIterator begin, InputIterator end) :
		std::basic_string<TCHAR>(begin, end) {};

	String& append(const basic_string<TCHAR>& str) {
		return *(String *)&std::basic_string<TCHAR>::append(str);
	}
	
	String& append(const TCHAR *s) {
		return *(String *)&std::basic_string<TCHAR>::append(s);
	}
	
	String& append(size_t n, TCHAR c) {
		return *(String *)&std::basic_string<TCHAR>::append(n, c);
	}
	
	String& append(TCHAR c) {
		return *(String *)&std::basic_string<TCHAR>::append(1, c);
	}
	
	String& operator+=(const basic_string<TCHAR>& str) {
		return *(String*)&std::basic_string<TCHAR>::operator+=(str);
	}
	String& operator+=(const TCHAR* s) {
		return *(String*)&std::basic_string<TCHAR>::operator+=(s);
	}
	String& operator+=(TCHAR c) {
		return *(String*)&std::basic_string<TCHAR>::operator+=(c);
	}
		
	String substr(size_t pos = 0, size_t n = npos) const {
		return String(*this, _M_check(pos, "String::substr"), n);
	}

	String substring(size_t beginIndex, size_t endIndex) const {
		return String(*this, _M_check(beginIndex, "String::substring"),
					  _M_check(endIndex, "String::substring")-beginIndex);
	}
	
	String substring(size_t beginIndex) const {
		return String(*this, _M_check(beginIndex, "String::substring"), npos);
	}	

	String trim() const {
		int beginIndex, endIndex;
		int length = size();
		for (beginIndex = 0; beginIndex < length; beginIndex++) {
			if (at(beginIndex) > 0x20) {
				for (endIndex = length; at(endIndex-1) <= 0x20; endIndex--);
				return String(*this, beginIndex, endIndex-beginIndex);
			}
		}
		return String(_T(""));
	}

	bool equals(const basic_string<TCHAR>& str) const {
		return compare(str) ? false : true;
	}	
	bool equals(const TCHAR* s) const {
		return compare(s) ? false : true;
	}
					
	bool startsWith(const TCHAR *prefix) const {
		size_t prefix_len = _tcslen(prefix);
		if (prefix_len > size()) return false;
		return compare(0, prefix_len, prefix) ? false : true;
	}

	bool startsWith(const basic_string<TCHAR>& prefix) const {
		size_t prefix_len = prefix.size();
		if (prefix_len > size()) return false;
		return compare(0, prefix_len, prefix) ? false : true;
	}

	bool endsWith(const TCHAR *suffix) const {
		size_t len = size();
		size_t suffix_len = _tcslen(suffix);
		if (suffix_len > len) return false;
		return compare(len-suffix_len, suffix_len, suffix) ? false : true;
	}

	bool endsWith(const basic_string<TCHAR>& suffix) const {
		size_t len = size();
		size_t suffix_len = suffix.size();
		if (suffix_len > len) return false;
		return compare(len-suffix_len, suffix_len, suffix) ? false : true;
	}

	const TCHAR& charAt(size_t pos) const { return at(pos); }
	
	const TCHAR& charAt(size_t pos) { return at(pos); }

	void split(TCHAR delim, StringList &result) {
		int pos = 0, i = 0;
		String token;
		for (; i < size(); i++) {
			if (at(i) == delim) {
				result.push_back(substr(pos, i-pos));
				pos = i+1;
			}
		}
		if (i-pos > 0)
			result.push_back(substr(pos, i-pos));
	}

	void casefold() {
		for (int i = 0; i < size(); i++) {
			at(i) = cl_tcasefold(at(i));
		}
	}
	
protected:
	size_type
	_M_check(size_type __pos, const char* __s) const
	{
		if (__pos > this->size())
			__throw_out_of_range(__N(__s));
		return __pos;
	}
	
};

inline String operator+ (TCHAR lhs, const String& rhs) {
	return String(lhs).append(rhs);
}

inline String operator+ (const TCHAR* lhs, const String& rhs) {
	return String(lhs).append(rhs);
}

inline String operator+ (const String& lhs, const TCHAR* rhs) {
	return String(lhs).append(rhs);
}

inline String operator+ (const String& lhs, TCHAR rhs) {
	return String(lhs).append(rhs);
}

inline String operator+ (const String& lhs, const String& rhs) {
	return String(lhs).append(rhs);
}

#endif
